#include "data.h"
#include "common.h"

namespace harkio {
using namespace std;

/**********************************************************************
 　Data 基底クラスメンバ関数
 **********************************************************************/

int Data::getClassSize() {
	int size = 0;
	for(unsigned int i=0; i < chunk.size(); i++) {
		size += chunk[i]->getClassSize();
	}
	size += strlen(getIdentifierFromBase()); // 自分自身のサイズを足す
//	size +  sizeof(size); Dataにはサイズがない

	return size;
}

void Data::addChunk(Chunk *chunk) {
	addChunkCheck(chunk);
	this->chunk.push_back(chunk);
}

vector<Chunk*> Data::getChunks() {
	return chunk;
}

void Data::setSize() {
	for(unsigned int i=0; i < chunk.size(); i++) {
		chunk[i]->setSize();
	}
}

/**********************************************************************
 　HGTF 派生クラスメンバ関数
 **********************************************************************/

bool HGTF::isComplete() {

	FMT *fmt = getChunk<FMT>();
	if (fmt == NULL) {
		HARKIO_SET_ERROR("Can't find FMT Chunk.");
		return false;
	}
	SPEC *spec = getChunk<SPEC>();
	if (spec == NULL) {
		HARKIO_SET_ERROR("Can't find SPEC Chunk.");
		return false;
	}
	int fid  = fmt->getFormatId();
	switch(fid) {
		case(FMT::M2PG):
		{
			M2PG*m2pg = getChunk<M2PG>();
			if (m2pg == NULL) {
				HARKIO_SET_ERROR("Can't find M2PG Chunk.");
				return false;
			}
			if(m2pg->isComplete() == false) {
				return false;
			}
		}
		break;
		case(FMT::GTF):
		{
			GTF*gtf = getChunk<GTF>();
			if (gtf == NULL) {
				HARKIO_SET_ERROR("Can't find GTF Chunk.");
				return false;
			}
			if(gtf->isComplete() == false) {
				return false;
			}
		}
		break;
		case(FMT::SM):
		{
			SM* sm = getChunk<SM>();
			if (sm == NULL) {
				HARKIO_SET_ERROR("Can't find SM Chunk.");
				return false;
			}
			if(sm->isComplete() == false) {
				return false;
			}
		}
		break;
		default:
			HARKIO_THROW_EXCEPTION(string("Unknown format id:") + boost::lexical_cast<string>(fid));
		break;
	}
	return true;
}

void HGTF::addChunkCheck(Chunk *chunk) {
	if (dynamic_cast<FMT*>(chunk)) {
		if (this->chunk.size()) { // chunk[0] をFTMにする
			HARKIO_THROW_EXCEPTION("FMT chunk must be first chunk.");
		}
	} else if(dynamic_cast<M2PG*>(chunk)) {
		// フォーマットIDの確認
		FMT *fmt = dynamic_cast<FMT*>(this->chunk[0]);
		if (fmt->getFormatId() != FMT::M2PG) {
			HARKIO_THROW_EXCEPTION("M2PG Chunk differed from FMT's formatID.");
		}
	} else if(dynamic_cast<GTF*>(chunk)) {
		// フォーマットIDの確認
		FMT *fmt = dynamic_cast<FMT*>(this->chunk[0]);
		if (fmt->getFormatId() != FMT::GTF) {
			HARKIO_THROW_EXCEPTION("GTF Chunk differed from FMT's formatID.");
		}
	} else if(dynamic_cast<SM*>(chunk)) {
		// フォーマットIDの確認
		FMT *fmt = dynamic_cast<FMT*>(this->chunk[0]);
		if (fmt->getFormatId() != FMT::SM) {
			HARKIO_THROW_EXCEPTION("SM Chunk differed from FMT's formatID.");
		}
	} else if (dynamic_cast<SPEC*>(chunk)) {
		return;
	} else {
		HARKIO_THROW_EXCEPTION(string("Unknown Chunk :") + typeid(chunk).name());
	}
}

void HGTF::writeFile(ofstream &fs) {

	// HGTF の識別子の書き込み
	fs.write((char*)getIdentifier(), strlen(getIdentifier()));
	if(fs.fail()) {
		HARKIO_THROW_ERROR("Can't write HGTF's data.");
	}

	// 包含データの書き込み関数呼び出し
	unsigned int chunk_count=0;
	for(chunk_count=0; chunk_count < chunk.size(); chunk_count++) {
		chunk[chunk_count]->writeFile(fs);
	}
	// チャンクの個数チェック
	if (chunk_count !=3) {
		HARKIO_THROW_ERROR("Count of Chunk wasn't complete.");
	}
	return;
}

void HGTF::readFile(ifstream &fs) {
	const char* identifier = getIdentifier();
	char cha_buf[strlen(identifier) + 1];

	// HGTFの識別子チェック
	memset(cha_buf, '\0', sizeof(cha_buf));
	fs.read((char*)cha_buf, strlen(identifier));
	if(fs.fail() || string(cha_buf) != identifier) {
		HARKIO_THROW_ERROR(string(identifier) + string(" was't HGTF's identifier."));
	}

	// FMT Chunk生成
	FMT *fmt = new FMT();
	addChunk(fmt);
	// FMT の内容を読み込み
	fmt->readFile(fs);

	// FormatIDに応じて、各チャンクの生成＆読み込み
	int format_id = fmt->getFormatId();
	switch(format_id) {
	case(FMT::M2PG):
	{
		M2PG *m2pg = new M2PG(fmt);
		addChunk(m2pg);
		m2pg->readFile(fs);
		break;
	}
	case(FMT::GTF):
	{
		GTF *gtf = new GTF(fmt);
		addChunk(gtf);
		gtf->readFile(fs);
		break;
	}
	case(FMT::SM):
	{
		SM *sm = new SM(fmt);
		addChunk(sm);
		sm->readFile(fs);
		break;
	}
	default:
		HARKIO_THROW_ERROR(string("Unknown format id :") + boost::lexical_cast<string>(format_id));
		break;
	}

	// SPEC Chunk生成
	SPEC *spec = new SPEC();
	addChunk(spec);
	// SPEC の内容を読み込み
	spec->readFile(fs);

	return;
}

} //
