#ifndef TEXTDATA_H
#define TEXTDATA_H
#include <iostream>
#include <boost/regex.hpp>
#include <boost/format.hpp>
#include "common.h"

namespace harkio {

template <typename T=float>
class UISrc {
public:
	T x;
	T y;
	T z;
	std::string implement_file;	//!srcinfテキストに埋め込まれたfltファイル名。fltファイル名が相対パスの時はフルパスに変換されている
	int cut_start;	//対象区間の開始　! srcinfファイルのfltファイル名の後ろの数 
	int cut_end;	//対象区間の終了
	UISrc() {
		x=0; y=0; z=0;
	}
};

template <typename T=float>
class UIMic{
public:
	T x;
	T y;
	T z;
	UIMic() {
		x=0; y=0; z=0;
	}
};

template <typename T=float>
class UINoise{
public:
	T x;
	T y;
	T z;
	UINoise() {
		x=0; y=0; z=0;
	}
};

class TextBase {
protected:
	int major_version;
	int minor_version;
	std::string reading_file; // テキスト自体のファイル名。このテキストを基準とする相対パスを絶対パスに変換する場合に参照する
public:
	TextBase() {}
	virtual ~TextBase() {}
	virtual void read(std::istream &is) = 0;
//	virtual void readText(std::string text) = 0;
//	virtual void readFile(std::string file) =0;

	void setVersion(int major, int minor) {
		major_version = major;
		minor_version = minor;
	}
	void version(int *major, int *minor) {
		*major = major_version;
		*minor = minor_version;
	}
	void setReadingFile(std::string file) {
		reading_file = file;
	}

	// ストリーム読み飛ばし
	void skipComment(std::istream& is) {
		std::string line;
		boost::regex regexp;
		std::streampos pos;
		pos = is.tellg();

		while(getline(is, line)) {
			CHOMP(line)
			regexp = "(?:^#.*|^$)";
			if (regex_search(line, regexp)) {
				pos = is.tellg();
				continue;
			} else {
				is.seekg(pos);
				break;
			}
		}
	}
	virtual bool checkIdentifier(std::istream& is) = 0;

	// 先頭の識別子だけを確認する
	bool checkIdentifier(std::istream& is, std::string pattern) {
		is.seekg(0); // 必ずもどしておく

		boost::regex regexp;
		std::string line;

		// コメント読み飛ばし
		skipComment(is);

		// ヘッダ
		if (getline(is, line)) {
			CHOMP(line)
			regexp = pattern + "\\s*(?:#+.*|)";
			if (!regex_search(line, regexp)) {
				return false;
			}
			return true;
		} else {
			return false;
		}
	}
	void readText(std::string text) {
		std::stringstream ss;
		ss << text;
		return read(ss);
	}
	void readFile(std::string file) {
		reading_file = file;
		std::ifstream ifs(file.c_str());
		if(ifs.fail())
		{
			std::string str = std::string("Can't Open file:") + file;
			HARKIO_THROW_ERROR(str);
		}
		return read(ifs);
	}

};

class ImpulseResponseListFile : public TextBase{
public:
	// Version
	typedef enum {V1_0 // 初期バージョン
	} VERSION;
	const static short versions[][2]; // [X][0] = Major [X][1] = Minor, 添字Xはenum に一致する
	std::vector<UISrc<float> > src_list;

	// 読み込むファイルが、このライブラリが扱える最新のバージョン以下か？を確認する
private:
	void underLatestVersion();
public:
	//読み込んだファイルが、少なくとも指定バージョン以上のファイルか？を確認する
	//expがtrueの時は、例外が発生する
	bool atLeastVersion(VERSION enum_index, bool exp=true) {
		short least_major = versions[enum_index][0];
		short least_minor = versions[enum_index][1];
		int ret;
		if (major_version < least_major || (major_version == least_major && minor_version < least_minor)) {
			ret = false;
		} else {
			ret = true;
		}
		if (!exp || ret) {
			return ret;
		} else {
			std::string error = (boost::format("ImpulseResponseListFile: %1%.%2% is less than %3%.%4%")
			% major_version % minor_version % least_major % least_minor).str();
			HARKIO_THROW_ERROR(error);
		}
	}

	std::vector<UISrc<float> >* srcLists() {
		return &src_list;
	}
	ImpulseResponseListFile() {}
	virtual ~ImpulseResponseListFile() {}
	virtual void read(std::istream& is);
	virtual bool checkIdentifier(std::istream& is) {
		return TextBase::checkIdentifier(is, "ImpulseResponse-ListFile");
	}
};

class TSPListFile : public TextBase{
	int sync_avg;
	std::string original_impulse;
	int tsp_offset;
	int signal_max;
public:
	// Version
	typedef enum {V1_0 // 初期バージョン
		, V1_1	// ChunnelList 項目の追加
	} VERSION;
	const static short versions[][2]; // [X][0] = Major [X][1] = Minor, 添字Xはenum に一致する

	std::vector<UISrc<float> > src_list;
	std::vector<int> channel_list;

	// 読み込むファイルが、このライブラリが扱える最新のバージョン以下か？を確認する
private:
	void underLatestVersion();
public:
	//読み込んだファイルが、少なくとも指定バージョン以上のファイルか？を確認する
	//expがtrueの時は、例外が発生する
	bool atLeastVersion(VERSION enum_index, bool exp=true) {
		short least_major = versions[enum_index][0];
		short least_minor = versions[enum_index][1];
		int ret;
		if (major_version < least_major || (major_version == least_major && minor_version < least_minor)) {
			ret = false;
		} else {
			ret = true;
		}
		if (!exp || ret) {
			return ret;
		} else {
			std::string error = (boost::format("TSPListFile version: %1%.%2% is less than %3%.%4%")
			% major_version % minor_version % least_major % least_minor).str();
			HARKIO_THROW_ERROR(error);
		}
	}

	std::vector<UISrc<float> >* srcLists() {
		return &src_list;
	}
	void setSyncAvg(int value) {
		sync_avg = value;
	}
	void setOriginalImpulse(std::string value) {
		original_impulse = value;
	}
	void setTSPOffset(int value) {
		tsp_offset = value;
	}
	void setSignalMax(int value) {
		signal_max = value;
	}
	int syncAvg() {
		return sync_avg;
	}
	std::string originalImpulse() {
		return original_impulse;
	}
	int tspOffset() {
		return tsp_offset;
	}
	int signalMax() {
		return signal_max;
	}

	TSPListFile() {}
	virtual ~TSPListFile() {}
	virtual void read(std::istream& is);
	virtual bool checkIdentifier(std::istream& is) {
		return TextBase::checkIdentifier(is, "TSP-ListFile");
	}
};

class MICARYLocationFile : public TextBase{
public:
	// Version
	typedef enum {V1_0 // 初期バージョン
	} VERSION;
	const static short versions[][2]; // [X][0] = Major [X][1] = Minor, 添字Xはenum に一致する
	std::vector<UIMic<float> > mic_lists;

	// 読み込むファイルが、このライブラリが扱える最新のバージョン以下か？を確認する
private:
	void underLatestVersion();
public:
	//読み込んだファイルが、少なくとも指定バージョン以上のファイルか？を確認する
	//expがtrueの時は、例外が発生する
	bool atLeastVersion(VERSION enum_index, bool exp=true) {
		short least_major = versions[enum_index][0];
		short least_minor = versions[enum_index][1];
		int ret;
		if (major_version < least_major || (major_version == least_major && minor_version < least_minor)) {
			ret = false;
		} else {
			ret = true;
		}
		if (!exp || ret) {
			return ret;
		} else {
			std::string error = (boost::format("MICARYLocationFile: %1%.%2% is less than %3%.%4%")
			% major_version % minor_version % least_major % least_minor).str();
			HARKIO_THROW_ERROR(error);
		}
	}

	MICARYLocationFile() {}
	virtual ~MICARYLocationFile() {}
	virtual void read(std::istream& is);
	virtual bool checkIdentifier(std::istream& is) {
		return TextBase::checkIdentifier(is, "MICARY-LocationFile");
	}
};

class NoiseLocationFile : public TextBase{
public:
	// Version
	typedef enum {V1_0 // 初期バージョン
	} VERSION;
	const static short versions[][2]; // [X][0] = Major [X][1] = Minor, 添字Xはenum に一致する
	std::vector<UINoise<float> > noise_lists;

	// 読み込むファイルが、このライブラリが扱える最新のバージョン以下か？を確認する
private:
	void underLatestVersion();
public:
	//読み込んだファイルが、少なくとも指定バージョン以上のファイルか？を確認する
	//expがtrueの時は、例外が発生する
	bool atLeastVersion(VERSION enum_index, bool exp=true) {
		short least_major = versions[enum_index][0];
		short least_minor = versions[enum_index][1];
		int ret;
		if (major_version < least_major || (major_version == least_major && minor_version < least_minor)) {
			ret = false;
		} else {
			ret = true;
		}
		if (!exp || ret) {
			return ret;
		} else {
			std::string error = (boost::format("NoiseLocationFile: %1%.%2% is less than %3%.%4%")
			% major_version % minor_version % least_major % least_minor).str();
			HARKIO_THROW_ERROR(error);
		}
	}

	NoiseLocationFile() {}
	virtual ~NoiseLocationFile() {}
	virtual void read(std::istream& is);
	virtual bool checkIdentifier(std::istream& is) {
		return TextBase::checkIdentifier(is, "Noise-LocationFile");
	}
};

class TextData {
public:
	class TYPE{
		public: enum {
		 UNDEFINE
		,ImpulseResponseListFile
		,TSPListFile
		,MICARYLocationFile
		,NoiseLocationFile
	};};

	TextData() {}
	virtual ~TextData() {}


	TextBase* readText(std::string text) {
		std::stringstream ss;
		ss << text;
		return read(ss);
	}
	TextBase* readFile(std::string file) {
		std::ifstream ifs(file.c_str());
		if(ifs.fail())
		{
			std::string str = std::string("Can't Open file:") + file;
			HARKIO_THROW_ERROR(str);
		}
		return read(ifs, file);
	}
	TextBase* read(std::istream& is, std::string fname="") {

		TextBase* text;

		int type = checkIdentifier(is);
		switch(type) {
		case TYPE::ImpulseResponseListFile:
			text = new ImpulseResponseListFile();
			break;
		case TYPE::TSPListFile:
			text = new TSPListFile();
			break;
		case TYPE::MICARYLocationFile:
			text = new MICARYLocationFile();
			break;
		case TYPE::NoiseLocationFile:
			text = new NoiseLocationFile();
			break;
		default:
			text = NULL;
			HARKIO_THROW_ERROR("Unknown format file");
			break;
		}
		text->setReadingFile(fname);
		text->read(is);
		return text;
	}

	int checkFileIdentifier(std::string file) {
		std::ifstream ifs(file.c_str());
		if(ifs.fail())
		{
			std::string str = std::string("Can't Open file:") + file;
			HARKIO_THROW_ERROR(str);
		}
		return checkIdentifier(ifs);
	}
	int checkTextIdentifier(std::string text) {
		std::stringstream ss;
		ss << text;
//		read(ss);
		return checkIdentifier(ss);
	}
	int checkIdentifier(std::istream& is) {
		bool ret;
		TextBase* text;

		text = new ImpulseResponseListFile();
		ret = text->checkIdentifier(is);
		delete text;
		if (ret) {
			return TYPE::ImpulseResponseListFile;
		}

		text = new TSPListFile();
		ret = text->checkIdentifier(is);
		if (ret) {
			return TYPE::TSPListFile;
		}

		text = new MICARYLocationFile();
		ret = text->checkIdentifier(is);
		if (ret) {
			return TYPE::MICARYLocationFile;
		}

		text = new NoiseLocationFile();
		ret = text->checkIdentifier(is);
		if (ret) {
			return TYPE::NoiseLocationFile;
		}
		return TYPE::UNDEFINE;
	}
};

} //namespace
#endif
